from .define import Dimension, Range

DIMENSION_ID_OVERWORLD = 0
DIMENSION_ID_NETHER = 1
DIMENSION_ID_END = 2

DIMENSION_OVERWORLD = Dimension(DIMENSION_ID_OVERWORLD)
DIMENSION_NETHER = Dimension(DIMENSION_ID_NETHER)
DIMENSION_END = Dimension(DIMENSION_ID_END)

RANGE_OVERWORLD = DIMENSION_OVERWORLD.range()
RANGE_NETHER = DIMENSION_NETHER.range()
RANGE_END = DIMENSION_END.range()
RANGE_INVALID = Range(0, -1)
